from src.usdm_excel.option_manager import *

def test_create():
  object = OptionManager()
  assert len(object.items.keys()) == 0
  assert object.items == {}

def test_set():
  option_manager.items = {}
  option_manager.set('fred', 'value')
  assert len(option_manager.items.keys()) == 1
  assert option_manager.items['fred'] == 'value'

def test_get():
  option_manager.items = {}
  option_manager.items['fred'] = 'value'
  assert option_manager.get('fred') == 'value'

def test_clear():
  option_manager.items = {}
  option_manager.items['fred'] = 'value'
  assert len(option_manager.items.keys()) == 1
  option_manager.clear()
  assert len(option_manager.items.keys()) == 0

def test_options():
  option_manager.set(Options.PREVIOUS_NEXT, PrevNextOption.NONE)
  assert option_manager.get(Options.PREVIOUS_NEXT) == PrevNextOption.NONE.value
  option_manager.set(Options.ROOT, RootOption.API_COMPLIANT)
  assert option_manager.get(Options.ROOT) == RootOption.API_COMPLIANT.value
  option_manager.set(Options.DESCRIPTION, 'Some text')
  assert option_manager.get(Options.DESCRIPTION) == 'Some text'


from typing import List, Dict, Union, Literal
from .api_base_model import ApiBaseModel
from .timing import Timing

class ScheduledInstance(ApiBaseModel):
    scheduledInstanceId: str
    scheduledInstanceType: Literal['ACTIVITY', 'DECISION']
    scheduleTimelineExitId: Union[str, None] = None
    scheduledInstanceTimings: List[Timing] = []
    scheduledInstanceTimelineId: Union[str, None] = None
    defaultConditionId: Union[str, None] = None
    epochId: Union[str, None] = None

class ScheduledActivityInstance(ScheduledInstance):
    activityIds: List[str] = []
    scheduledActivityInstanceEncounterId: Union[str, None] = None

class ScheduledDecisionInstance(ScheduledInstance):
    conditionAssignments: List[List] = []


from usdm_excel.base_sheet import BaseSheet
from usdm_excel.study_soa_sheet.soa_column_rows import SoAColumnRows
from usdm_excel.study_soa_sheet.timepoint import Timepoint
from usdm_excel.id_manager import id_manager
import pandas as pd

class Timepoints():
  
  def __init__(self, parent):
    #super().__init__(sheet)
    self.parent = parent
    self.items = []
    self.map = {}
    self.activity_names = []
    self._build_activities()
    self._build_timepoints()
    self._set_to_timing_refs()

  def item_at(self, key):
    return self.map[key]

  def insert_at(self, insert_at_index, type, value, cycle, reference=None):
    timepoint = Timepoint(self.parent, self.activity_names, None, type, value, cycle, additional=True)
    timepoint.reference = reference
    self.items.insert(insert_at_index, timepoint)
    return timepoint
  
  def set_condition_refs(self):
    for item in self.items:
      condition = item.usdm_timepoint
      if condition.scheduledInstanceType == 'DECISION':
        condition_instance = self.items[item.reference].usdm_timepoint
        text = item.timing_value
        if text == "":
          text = "default, no condition set"
        condition.conditionAssignments.append([text, condition_instance.scheduledInstanceId])
    previous_item = None
    for item in self.items:
      if previous_item == None:
        previous_item = item        
        continue
      previous_condition = previous_item.usdm_timepoint
      if previous_condition.scheduledInstanceType == 'DECISION':
        current_instance = item.usdm_timepoint
        previous_condition.conditionAssignments.append(["default", current_instance.scheduledInstanceId])        
      previous_item = item        

  def _build_timepoints(self):    
    for col_index in range(self.parent.sheet.shape[1]):
      if col_index >= SoAColumnRows.FIRST_VISIT_COL:
        record = Timepoint(self.parent, self.activity_names, col_index)
        self.items.append(record)
        self.map[record.key()] = record

  def _build_activities(self):    
    for row_index, col_def in self.parent.sheet.iterrows():
      if row_index >= SoAColumnRows.FIRST_ACTIVITY_ROW:
        activity = self.parent.read_cell(row_index, SoAColumnRows.CHILD_ACTIVITY_COL)
        if not activity == "":
          self.activity_names.append(activity)

  def _set_to_timing_refs(self):    
    for item in self.items:
      from_instance = item.usdm_timepoint
      from_timing = from_instance.scheduledInstanceTimings[0]
      from_timing_type = from_timing.timingType.code
      if from_timing_type == "ANCHOR":
        continue
      to_instance = self.items[item.reference].usdm_timepoint
      from_timing.relativeToScheduledInstanceId = to_instance.scheduledInstanceId


from usdm_excel.base_sheet import BaseSheet
from usdm_excel.cross_ref import cross_references
from usdm_excel.id_manager import id_manager
from usdm_model.study_arm import StudyArm
import traceback

class StudyDesignArmSheet(BaseSheet):

  def __init__(self, file_path):
    try:
      super().__init__(file_path=file_path, sheet_name='studyDesignArms')
      self.items = []
      for index, row in self.sheet.iterrows():
        name = self.read_cell_by_name(index, 'studyArmName')
        description = self.read_description_by_name(index, 'studyArmDescription')
        arm_type = self.read_cdisc_klass_attribute_cell_by_name('StudyArm', 'studyArmType', index, 'studyArmType')
        arm_origin_description = self.read_description_by_name(index, 'studyArmDataOriginDescription')
        arm_origin_type = self.read_cdisc_klass_attribute_cell_by_name('StudyArm', 'studyArmDataOriginType', index, 'studyArmDataOriginType')
        try:
          item = StudyArm(
            studyArmId=id_manager.build_id(StudyArm), 
            studyArmName=name,
            studyArmDescription=description,
            studyArmType=arm_type,
            studyArmDataOriginDescription=arm_origin_description,
            studyArmDataOriginType=arm_origin_type
          )
        except Exception as e:
          self._general_error(f"Failed to create StudyArm object, exception {e}")
        else:
          self.items.append(item)
          cross_references.add(name, item)     
    except Exception as e:
      self._general_error(f"Exception [{e}] raised reading sheet.")
      self._traceback(f"{traceback.format_exc()}")




from usdm_excel.id_manager import id_manager
from usdm_model.code import Code
import json
import os

class ISO3166():

  def __init__(self):
    f = open(os.path.join(os.path.dirname(__file__), 'data', 'iso_3166.json'))
    self.db = json.load(f)

  def code(self, code):
    code, decode = self._get_decode(code)
    return Code(codeId=id_manager.build_id(Code), code=code, codeSystem='ISO 3166 1 alpha3', codeSystemVersion='2020-08', decode=decode)

  def _get_decode(self, code):
    if len(code) == 2:
      field = 'alpha-2'
    else:
      field = 'alpha-3'
    entry = next((item for item in self.db if item[field] == code), None)
    if entry == None:
      return 'DNK', 'Denmark'
    else:
      return entry['alpha-3'], entry['name']


from typing import List, Union
from .api_base_model import ApiBaseModel
from .code import Code

from uuid import UUID

class AliasCode(ApiBaseModel):
  aliasCodeId: str
  standardCode: Code
  standardCodeAliases: List[Code] = []



from typing import Union
from .api_base_model import ApiBaseModel
from .code import Code

class StudyArm(ApiBaseModel):
  studyArmId: str
  studyArmName: str
  studyArmDescription: Union[str, None] = None
  studyArmType: Code
  studyArmDataOriginDescription: str
  studyArmDataOriginType: Code



from usdm_excel.study_soa_sheet.soa_column_rows import SoAColumnRows
from usdm_excel.base_sheet import BaseSheet
from usdm_excel.id_manager import id_manager
from usdm_excel.study_soa_sheet.cycles import Cycles
from usdm_excel.study_soa_sheet.timepoints import Timepoints
from usdm_excel.study_soa_sheet.timepoint import Timepoint
from usdm_excel.study_soa_sheet.activities import Activities
from usdm_model.scheduled_instance import ScheduledActivityInstance, ScheduledDecisionInstance
from usdm_model.schedule_timeline import ScheduleTimeline
from usdm_model.schedule_timeline_exit import ScheduleTimelineExit

import traceback

class StudySoASheet(BaseSheet):

  NAME_ROW = 0
  DESCRIPTION_ROW = 1
  CONDITION_ROW = 2
  PARAMS_DATA_COL = 1

  def __init__(self, file_path, sheet_name, main=False):
    try:
      super().__init__(file_path=file_path, sheet_name=sheet_name, header=None)
      self.name = ""
      self.description = ""
      self.condition = ""
      self.timeline = None
      self.main_timeline = main
      self.encounters = []
      self.activities = []
      self.timelines = []
      self.biomedical_concept_surrogates = []
      self.biomedical_concepts = []
      self._process_sheet()
      self._raw_cycles = Cycles(self)
      self._raw_timepoints = Timepoints(self)
      #self._raw_encounters = Encounters(self)
      self._raw_activities = Activities(self)

      self._link_instance_to_activities()
      self._insert_cycles_into_timeline()
      self._raw_timepoints.set_condition_refs()

      instances = []

      for item in self._raw_activities.items:
        self.activities.append(item.usdm_activity)
        self.biomedical_concept_surrogates += item.usdm_biomedical_concept_surrogates
        self.biomedical_concepts += item.usdm_biomedical_concepts
      self.double_link(self.activities, 'activityId', 'previousActivityId', 'nextActivityId')
      
      prev_instance = None
      for raw_timepoint in self._raw_timepoints.items:
        instance = raw_timepoint.usdm_timepoint
        instance.defaultConditionId = None
        instances.append(instance)
        if prev_instance is not None:
          prev_instance.defaultConditionId = instance.scheduledInstanceId
        prev_instance = instance
      exit = self._add_exit()
      self.timeline = self._add_timeline(self.name, self.description, self.condition, instances, exit)

    except Exception as e:
      print(f"Exception [{e}] raised reading sheet")
      self._general_error(f"Exception [{e}] raised reading sheet")
      self._traceback(f"{traceback.format_exc()}")

  def _process_sheet(self):
    for rindex in range(self.NAME_ROW, self.CONDITION_ROW + 1):
      if rindex == self.NAME_ROW:
        self.name = self.read_cell(rindex, self.PARAMS_DATA_COL)
      elif rindex == self.DESCRIPTION_ROW:
        self.description = self.read_cell(rindex, self.PARAMS_DATA_COL)
      elif rindex == self.CONDITION_ROW:
        self.condition = self.read_cell(rindex, self.PARAMS_DATA_COL)
      else:
        pass

  def _add_exit(self):
    return ScheduleTimelineExit(scheduleTimelineExitId=id_manager.build_id(ScheduleTimelineExit))

  def _add_timeline(self, name, description, condition, instances, exit):
    return ScheduleTimeline(
      scheduleTimelineId=id_manager.build_id(ScheduleTimeline),
      mainTimeline=self.main_timeline,
      scheduleTimelineName=name,
      scheduleTimelineDescription=description,
      entryCondition=condition,
      scheduleTimelineEntryId=instances[0].scheduledInstanceId,
      scheduleTimelineExits=[exit],
      scheduleTimelineInstances=instances
    )

  def _link_instance_to_activities(self):
    for timepoint in self._raw_timepoints.items:
      if timepoint.has_activities:
        for activity_name, selected in timepoint.activity_map.items():
          if selected:
            activity = self._raw_activities.item_by_name(activity_name)
            timepoint.add_activity(activity)

  def _insert_cycles_into_timeline(self):
    cycle_offset = 0
    for cycle in self._raw_cycles.items:
      start_index = cycle.start_timepoint_index + cycle_offset
      self._raw_timepoints.insert_at(start_index, 'anchor', cycle.start, cycle.cycle)
      cycle_offset += 1

      end_index = cycle.end_timepoint_index + cycle_offset + 1
      self._raw_timepoints.insert_at(end_index, 'previous', cycle.period, None)
      cycle_offset += 1

      if cycle.end_rule != "":
        end_index = cycle.end_timepoint_index + cycle_offset + 1
        self._raw_timepoints.insert_at(end_index, 'condition', cycle.end_rule, None, start_index)
        cycle_offset += 1




from usdm_excel.study_soa_sheet.soa_column_rows import SoAColumnRows
from usdm_excel.id_manager import id_manager
import pandas as pd

class Cycle():
  
  def __init__(self, parent, col_index, cycle, timepoint_index):
    self.parent = parent
    self.col_index = col_index
    self.position_key = col_index - SoAColumnRows.FIRST_VISIT_COL
    self.cycle = cycle
    self.start_timepoint_index = timepoint_index
    self.end_timepoint_index = timepoint_index
    self.start = self.get_cycle_cell(SoAColumnRows.CYCLE_START_ROW, col_index)
    self.period = self.get_cycle_cell(SoAColumnRows.CYCLE_PERIOD_ROW, col_index)
    self.end_rule = self.get_cycle_cell(SoAColumnRows.CYCLE_END_RULE_ROW, col_index)

  def add_end(self, index):
    self.end_timepoint_index = index

  def get_cycle_cell(self, row_index, col_index):
    return self.parent.read_cell_empty(row_index, col_index, '-')




class ISO8601Duration():
  
  ZERO_DURATION = "PT0M"

  def __init__(self):
    pass

  def encode(self, value, unit):
    the_value = int(value.strip())
    unit_text = unit.strip()
    if unit_text.upper() in ['Y', 'YRS', 'YR', 'YEARS', 'YEAR']:
      return f"P{the_value}Y"
    if unit_text.upper() in ['MTHS', 'MTH', 'MONTHS', 'MONTH']:
      return f"P{the_value}M"
    if unit_text.upper() in ['W', 'WKS', 'WK', 'WEEKS', 'WEEK']:
      return f"P{the_value}W"
    if unit_text.upper() in ['D', 'DYS', 'DY', 'DAYS', 'DAY']:
      return f"P{the_value}D"
    if unit_text.upper() in ['H', 'HRS', 'HR', 'HOURS', 'HOUR']:
      return f"PT{the_value}H"
    if unit_text.upper() in ['M', 'MINS', 'MIN', 'MINUTES', 'MINUTE']:
      return f"PT{the_value}M"
    if unit_text.upper() in ['S', 'SECS', 'SEC', 'SECONDS', 'SECOND']:
      return f"PT{the_value}S"
    raise ValueError
  


import pytest
from src.usdm_excel.iso_8601_duration import ISO8601Duration

def test_encode(mocker):
  item = ISO8601Duration()
  assert item.encode("1", "y") == "P1Y"
  assert item.encode("2", "yeaR") == "P2Y"
  assert item.encode("2", "years") == "P2Y"
  assert item.encode("3", "yrs") == "P3Y"
  assert item.encode("3", "yr") == "P3Y"
  assert item.encode("1", "MTHS") == "P1M"
  assert item.encode("1", "MTH") == "P1M"
  assert item.encode("2", "mOnth") == "P2M"
  assert item.encode("2", "months") == "P2M"
  assert item.encode("12", "week") == "P12W"
  assert item.encode("12", "weeks") == "P12W"
  assert item.encode("13", "Wks") == "P13W"
  assert item.encode("13", "Wk") == "P13W"
  assert item.encode("1 ", "W") == "P1W"
  assert item.encode("1", "d") == "P1D"
  assert item.encode("12", "days") == "P12D"
  assert item.encode("12", "dys") == "P12D"
  assert item.encode("12", "dy") == "P12D"
  assert item.encode("13", "DAY") == "P13D"
  assert item.encode("1 ", "Hours") == "PT1H"
  assert item.encode("1 ", "Hour") == "PT1H"
  assert item.encode("1 ", "HRS") == "PT1H"
  assert item.encode("1 ", "HR") == "PT1H"
  assert item.encode(" 2", "H") == "PT2H"
  assert item.encode("  3  ", "H") == "PT3H"
  assert item.encode("1 ", "Mins") == "PT1M"
  assert item.encode("1 ", "MiN") == "PT1M"
  assert item.encode(" 2", "M") == "PT2M"
  assert item.encode("  322  ", "minutes") == "PT322M"
  assert item.encode("  322  ", "minute") == "PT322M"
  assert item.encode("1 ", "secS") == "PT1S"
  assert item.encode("1 ", "seC") == "PT1S"
  assert item.encode(" 2", "S") == "PT2S"
  assert item.encode("  322  ", "Seconds") == "PT322S"
  assert item.encode("  322  ", "Second") == "PT322S"
  with pytest.raises(Exception):
    item.encode("  322  ", "XSeconds")


from src.usdm_excel.errors.errors import Errors

def test_create():
  errors = Errors()
  assert errors.items == []

def test_add():
  errors = Errors()
  assert len(errors.items) == 0
  errors.add(sheet="My Sheet", row=1, column=99, message="XXXXX")
  assert len(errors.items) == 1
  assert errors.items[0].message == "XXXXX"
    
def test_count():
  errors = Errors()
  errors.add(sheet="My Sheet", row=1, column=99, message="XXXXX")
  errors.add(sheet="My Sheet", row=1, column=99, message="XXXXX")
  assert errors.count() == 2

def test_clear():
  errors = Errors()
  assert len(errors.items) == 0
  errors.add(sheet="My Sheet", row=1, column=99, message="XXXXX")
  assert len(errors.items) == 1
  errors.clear()
  assert len(errors.items) == 0

def test_dumo():
  errors = Errors()
  errors.add(sheet="My Sheet", row=1, column=99, message="XXXXX1")
  errors.add(sheet="My Sheet", row=2, column=100, message="XXXXX2")
  errors.add(sheet="My Sheet", row=3, column=101, message="XXXXX3", level=Errors.WARNING)
  errors.add(sheet="My Sheet", row=4, column=102, message="XXXXX4", level=Errors.INFO)
  errors.add(sheet="My Sheet", row=5, column=103, message="XXXXX5", level=Errors.DEBUG)
  assert (errors.dump(Errors.DEBUG)) == [
    {'sheet': 'My Sheet', 'row': 1, 'column': 99, 'message': 'XXXXX1', 'level': 'Error'},
    {'sheet': 'My Sheet', 'row': 2, 'column': 100, 'message': 'XXXXX2', 'level': 'Error'},
    {'sheet': 'My Sheet', 'row': 3, 'column': 101, 'message': 'XXXXX3', 'level': 'Warning'},
    {'sheet': 'My Sheet', 'row': 4, 'column': 102, 'message': 'XXXXX4', 'level': 'Info'},
    {'sheet': 'My Sheet', 'row': 5, 'column': 103, 'message': 'XXXXX5', 'level': 'Debug'}
  ]
  assert (errors.dump(Errors.ERROR)) == [
    {'sheet': 'My Sheet', 'row': 1, 'column': 99, 'message': 'XXXXX1', 'level': 'Error'},
    {'sheet': 'My Sheet', 'row': 2, 'column': 100, 'message': 'XXXXX2', 'level': 'Error'}
  ]
  assert (errors.dump(Errors.WARNING)) == [
    {'sheet': 'My Sheet', 'row': 1, 'column': 99, 'message': 'XXXXX1', 'level': 'Error'},
    {'sheet': 'My Sheet', 'row': 2, 'column': 100, 'message': 'XXXXX2', 'level': 'Error'},
    {'sheet': 'My Sheet', 'row': 3, 'column': 101, 'message': 'XXXXX3', 'level': 'Warning'}
  ]



from typing import List, Union
from .api_base_model import ApiBaseModel
from .code import Code
from .endpoint import Endpoint

class Objective(ApiBaseModel):
  objectiveId: str
  objectiveDescription: str
  objectiveLevel: Union[Code, None] = None
  objectiveEndpoints: List[Endpoint] = []


from typing import List, Union
from .api_base_model import ApiBaseModel
from .alias_code import AliasCode

class BiomedicalConceptCategory(ApiBaseModel):
  biomedicalConceptCategoryId: str
  bcCategoryChildIds: List[str] = []
  bcCategoryName: str
  bcCategoryDescription: Union[str, None] = None
  bcCategoryMemberIds: List[str] = []
  bcCategoryCode: Union[AliasCode, None] = None



from usdm_excel.base_sheet import BaseSheet
from usdm_excel.study_soa_sheet.soa_column_rows import SoAColumnRows
from usdm_excel.study_soa_sheet.timepoint_type import TimepointType
from usdm_excel.study_soa_sheet.window_type import WindowType
from usdm_excel.id_manager import id_manager
from usdm_excel.cross_ref import cross_references
from usdm_excel.cdisc_ct import CDISCCT
from usdm_model.timing import Timing
from usdm_model.scheduled_instance import ScheduledActivityInstance, ScheduledDecisionInstance

class Timepoint():
  
  def __init__(self, parent, activity_names, col_index, type="", value="", cycle=None, additional=False):
    self.parent = parent
    self.col_index = col_index
    if col_index == None:
      self._position_key = None
    else:
      self._position_key = col_index - SoAColumnRows.FIRST_VISIT_COL
    self.encounter_xref = ""
    self.has_activities = not additional
    self.has_encounter = False
    if not additional:
      self.encounter_xref, encounter_is_null = self.parent.read_cell_empty_legacy(SoAColumnRows.VISIT_LABEL_ROW, col_index)
      if not self.encounter_xref == "":
        self.has_encounter = True
      self.epoch = self.parent.read_cell_with_previous(SoAColumnRows.EPOCH_ROW, col_index, SoAColumnRows.FIRST_VISIT_COL)
      epoch_ref = cross_references.get(self.epoch)
    self.activities = []
    self.activity_map = {}
    self.__timepoint_type = None
    self.reference = None
    #self.timing_type = type
    #self.timing_value = value
    self.__window = None
    self.cycle = cycle
    if not additional:
      self._process_timepoint()
      self._add_activities(activity_names)
    self.usdm_timepoint = self._as_usdm()
    if self.has_encounter:
      encounter = cross_references.get(self.encounter_xref)
      self.usdm_timepoint.scheduledActivityInstanceEncounterId = encounter.encounterId
    if epoch_ref is not None:
      self.usdm_timepoint.epochId = epoch_ref.studyEpochId

  def key(self):
    return self._position_key
  
  def add_activity(self, activity):
    self.usdm_timepoint.activityIds.append(activity.usdm_activity.activityId)

  def _process_timepoint(self):
    self.__timepoint_type = TimepointType(self.parent, SoAColumnRows.TIMING_ROW, self.col_index)
    self.reference = self.col_index - SoAColumnRows.FIRST_VISIT_COL + self.__timepoint_type.relative_ref
    self.__window = WindowType(self.parent, SoAColumnRows.VISIT_WINDOW_ROW, self.col_index)

  def _as_usdm(self):
    instance = None
    if self.__timepoint_type.timing_type in ["anchor", "next", "previous", "cycle start"]:
      timing = self._to_timing()
      instance = ScheduledActivityInstance(
        scheduledInstanceId=id_manager.build_id(ScheduledActivityInstance),
        scheduledInstanceType='ACTIVITY',
        scheduleTimelineExitId=None,
        scheduledInstanceEncounterId=None,
        scheduledInstanceTimings=[timing],
        scheduledInstanceTimelineId=None,
        defaultConditionId=None,
        epochId=None,
        activityIds=[]
      )
      timing.relativeFromScheduledInstanceId = instance.scheduledInstanceId
    elif self.__timepoint_type.timing_type == "condition":
      instance = ScheduledDecisionInstance(
        scheduledInstanceId=id_manager.build_id(ScheduledActivityInstance),
        scheduledInstanceType='DECISION',
        scheduleTimelineExitId=None,
        scheduledInstanceEncounterId=None,
        scheduledInstanceTimings=[],
        scheduledInstanceTimelineId=None,
        defaultConditionId=None,
        conditionAssignments=[]
      )
    else:
      self.parent.general_warning(f"Unrecognized ScheduledInstance type: '{self.__timepoint_type.timing_type}'")
    return instance

  def _to_timing(self):
    type_code = {
      "ANCHOR": CDISCCT().code('C99901x3', 'Fixed Reference'),
      "PREVIOUS": CDISCCT().code('C99901x1', 'After'),
      "NEXT": CDISCCT().code('C99901x2', 'Before'),
      "CYCLE START": CDISCCT().code('C99901x3', 'Fixed Reference'),
    }
    return Timing(
      timingId=id_manager.build_id(Timing),
      timingType=type_code[self.__timepoint_type.timing_type.upper()],
      timingValue=self.__timepoint_type.value,
      timingDescription=self.__timepoint_type.description,
      timingRelativeToFrom=CDISCCT().code('C99900x1', 'Start to Start'),
      timingWindow=self.__window.description,
      timingWindowLower=self.__window.lower,
      timingWindowUpper=self.__window.upper,
      relativeFromScheduledInstanceId='',
      relativeToScheduledInstanceId=''
    )

  def get_relative_ref(self, part):
    if len(part) > 1:
      return int(part[1:])
    else:
      return 1

  def _add_activities(self, activity_names):
    for activity in activity_names:
      self.activity_map[activity] = False
    column = self.parent.sheet.iloc[:, self.col_index]
    row = 0
    for cell in column:
      if row >= SoAColumnRows.FIRST_ACTIVITY_ROW:
        activity = self.parent.read_cell(row, SoAColumnRows.CHILD_ACTIVITY_COL)
        if str(cell).upper() == "X":
          self.activity_map[activity] = True
      row += 1




from usdm_excel.id_manager import id_manager
from usdm_model.code import Code

class NCIt():

  def __init__(self):
    self.id_manager = id_manager

  def code(self, code, decode):
    return Code(codeId=id_manager.build_id(Code), code=code, codeSystem='NCI Thesaurus', codeSystemVersion='', decode=decode)




class IdManager():

  def __init__(self):
    self.id_index = {}
    self.clear()

  def clear(self):
    self.id_index = {
      'Address': 0,
      'Code': 0,
      'AliasCode': 0,
      'Organisation': 0,
      'StudyIdentifier': 0,
      'StudyEpoch': 0,
      'StudyArm': 0,
      'StudyCell': 0,
      'Entry': 0, 
      'ScheduleTimelineExit': 0, 
      'ScheduledActivityInstance': 0,
      'ScheduleTimeline': 0, 
      'Timing': 0, 
      'StudyDesign': 0, 
      'Study': 0, 
      'Activity': 0, 
      'Encounter': 0, 
      'BiomedicalConceptSurrogate': 0, 
      'BiomedicalConcept': 0,
      'BiomedicalConceptProperty': 0,
      'ResponseCode': 0,
      'InvestigationalIntervention': 0,
      'Indication': 0,
      'StudyDesignPopulation': 0,
      'Objective': 0,
      'Endpoint': 0,
      'IntercurrentEvent': 0,
      'Estimand': 0,
      'AnalysisPopulation': 0,
      'StudyProtocolVersion': 0,
      'Procedure': 0,
      'TransitionRule': 0,
      'StudyElement': 0
    }

  def build_id(self, klass):
    klass_name = str(klass.__name__)
    self.id_index[klass_name] += 1
    return "%s_%s" % (klass_name, self.id_index[klass_name])

id_manager = IdManager()




from typing import List, Union
from .api_base_model import ApiBaseModel
from .analysis_population import AnalysisPopulation
from .investigational_intervention import InvestigationalIntervention
from .endpoint import Endpoint
from .intercurrent_event import IntercurrentEvent
from uuid import UUID

class Estimand(ApiBaseModel):
  estimandId: str
  summaryMeasure: str
  analysisPopulation: AnalysisPopulation
  treatment: Union[str, None] = None
  variableOfInterest: Union[str, None] = None
  intercurrentEvents: List[IntercurrentEvent] = []



class CrossRef():

  def __init__(self):
    self.references = {}

  def clear(self):
    self.references = {}

  def add(self, key, value):
    if not key in self.references:
      self.references[key] = value
    else:
      pass # Need exception here

  def get(self, key):
    if key in self.references:
      return self.references[key]
    else:
      return None

cross_references = CrossRef()


from usdm_excel.base_sheet import BaseSheet
from usdm_excel.cross_ref import cross_references
from usdm_excel.id_manager import id_manager
import traceback
import pandas as pd
from usdm_model.objective import Objective
from usdm_model.endpoint import Endpoint

class StudyDesignObjectiveEndpointSheet(BaseSheet):

  def __init__(self, file_path):
    try:
      super().__init__(file_path=file_path, sheet_name='studyDesignOE')
      self.objectives = []
      current = None
      for index, row in self.sheet.iterrows():
        o_xref = self.read_cell_by_name(index, "objectiveXref")
        o_description = self.read_cell_by_name(index, 'objectiveDescription') # Note, dont use description read method, we need to know if really empty.
        e_xref = self.read_cell_by_name(index, "endpointXref")
        e_description = self.read_description_by_name(index, 'endpointDescription')
        e_p_description = self.read_description_by_name(index, 'endpointPurposeDescription')
        e_level = self.read_cdisc_klass_attribute_cell_by_name('Endpoint', 'endpointLevel', index, "endpointLevel")
        if not o_description == "":
          o_level = self.read_cdisc_klass_attribute_cell_by_name('Objective', 'objectiveLevel', index, "objectiveLevel")
          try:
            current = Objective(objectiveId=id_manager.build_id(Objective),
              objectiveDescription=o_description, 
              objectiveLevel=o_level,
              objectiveEndpoints=[]
            )
          except Exception as e:
            self._general_error(f"Failed to create Objective object, exception {e}")
          else:
            self.objectives.append(current)
            cross_references.add(o_xref, current.objectiveId)
        if current is not None:
          try:
            ep = Endpoint(endpointId=id_manager.build_id(Endpoint), 
              endpointDescription=e_description, 
              endpointPurposeDescription=e_p_description, 
              endpointLevel=e_level
            )  
          except Exception as e:
            self._general_error(f"Failed to create Endpoint object, exception {e}")
          else:
            current.objectiveEndpoints.append(ep)
            cross_references.add(e_xref, ep.endpointId)
        else:
          self._general_error("Failed to add Endpoint, no Objective set")

    except Exception as e:
      self._general_error(f"Exception [{e}] raised reading sheet.")
      self._traceback(f"{traceback.format_exc()}")




import json
import csv
from src.usdm_excel import USDMExcel

SAVE_ALL = False

def save_error_csv(file, contents):
  writer = csv.DictWriter(file, fieldnames=['sheet','row','column','message','level'])
  writer.writeheader()
  writer.writerows(contents)

def read_error_csv(file):
  reader = csv.DictReader(file)
  items = list(reader)
  for item in items:
    item['row'] = int(item['row'])
    item['column'] = int(item['column'])
  return items

def run_test(filename, save=False):
  print(f"RUN: S={save}")
  excel = USDMExcel(f"tests/integration_test_files/{filename}.xlsx")
  result = excel.to_json()
  errors = excel.errors()

  # Useful if you want to see the results.
  print(f"ERRORS: S={save} E={errors}")
  if save or SAVE_ALL:
    with open(f"tests/integration_test_files/{filename}.json", 'w', encoding='utf-8') as f:
      f.write(json.dumps(json.loads(result), indent=2))
    print(f"ERRORS: {errors}")
    with open(f"tests/integration_test_files/{filename}_errors.csv", 'w',newline='') as f:
      save_error_csv(f, errors) 
  
  with open(f"tests/integration_test_files/{filename}.json", 'r') as f:
    expected = json.dumps(json.load(f)) # Odd, but doing it for consistency of processing
  assert result == expected
  with open(f"tests/integration_test_files/{filename}_errors.csv", 'r') as f:
    expected = read_error_csv(f)
  assert errors == expected

def run_test_ne(filename, save=False):
  result = {}
  excel = USDMExcel(f"tests/integration_test_files/{filename}.xlsx")
  result['n'], result['e'] = excel.to_nodes_and_edges()
  for type in ['n', 'e']:

    # Useful if you want to see the results.
    if save or SAVE_ALL:
      with open(f"tests/integration_test_files/{filename}_{type}.json", 'w', encoding='utf-8') as f:
        f.write(json.dumps(result[type], indent=2))
    
    with open(f"tests/integration_test_files/{filename}_{type}.json", 'r') as f:
      expected = json.load(f)
    assert result[type] == expected

def test_simple_1():
  run_test('simple_1')

def test_simple_1_ne():
  run_test_ne('simple_1')

def test_config_1():
  run_test('config_1')

def test_config_2():
  run_test('config_2')

def test_no_activity_sheet():
  run_test('no_activity_sheet')

def test_address_comma():
  run_test('address')

def test_complex_1():
  run_test('complex_1')

def test_complex_1_ne():
  run_test_ne('complex_1')

def test_arms_epochs_1():
  run_test('arms_epochs')

def test_arms_epochs_1_ne():
  run_test_ne('arms_epochs')



from typing import Union
from .api_base_model import ApiBaseModel
from .code import Code
from .address import Address

class Organisation(ApiBaseModel):
  organisationId: str
  organisationIdentifierScheme: str
  organisationIdentifier: str
  organisationName: str
  organisationType: Code
  organizationLegalAddress: Union[Address, None] = None



from typing import Union
from uuid import UUID
from .api_base_model import ApiBaseModel
from .organisation import Organisation

class StudyIdentifier(ApiBaseModel):
  studyIdentifierId: str
  studyIdentifier: str
  studyIdentifierScope: Organisation



from typing import Union
from .api_base_model import ApiBaseModel
from .code import Code

class Procedure(ApiBaseModel):
  procedureId: str
  procedureType: str
  procedureName: str
  procedureDescription: Union[str, None] = None
  procedureCode: Code
  procedureIsConditional: bool
  procedureIsConditionalReason: Union[str, None] = None



from pydantic import BaseModel
import json
import enum
import datetime

def _serialize_as_json(obj):
  #print("OBJ:", type(obj))
  if isinstance(obj, enum.Enum):
    return obj.value
  elif isinstance(obj, datetime.date):
    return obj.isoformat()
  else:
    return obj.__dict__

def _serialize_as_json_with_type(obj):
  # Example, see https://stackoverflow.com/questions/10252010/serializing-class-instance-to-json
  if isinstance(obj, enum.Enum):
    return obj.value
  elif isinstance(obj, datetime.date):
    return obj.isoformat()
  else:
    result = obj.__dict__
    result['_type'] = obj.__class__.__name__
    return result

class ApiBaseModel(BaseModel):

  def to_json(self):
    return json.dumps(self, default=_serialize_as_json)

  def to_json_with_type(self):
    return json.dumps(self, default=_serialize_as_json_with_type)




import pytest
import pandas as pd

from src.usdm_excel.study_identifiers_sheet.study_identifiers_sheet import StudyIdentifiersSheet
from src.usdm_model.code import Code

def test_create(mocker):
  mock_id = mocker.patch("usdm_excel.id_manager.build_id")
  mock_id.side_effect=['Code_1', 'Org_1', 'Addr_1', 'Id_1', 'Code_2', 'Org_2', 'Addr_2', 'Id_2', 'Code_3', 'Org_3', 'Addr_3', 'Id_3']
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  data = [
    [ 'USGOV', 'CT-GOV', 'ClinicalTrials.gov', 'Study Registry', 'NCT12345678', 'line|city|district|state|postal_code|GBR' ],
    [ 'USGOV2', 'CT-GOV2', 'ClinicalTrials2.gov', 'Study Registry', 'NCT12345679', 'line2,city2,district2,state2,postal_code2,FRA' ],
    [ 'USGOV3', 'CT-GOV3', 'ClinicalTrials3.gov', 'Study Registry', 'NCT123456710', 'line3,city3,district3,state3,postal_code3,FR' ]
  ]
  mock_read = mocker.patch("pandas.read_excel")
  mock_read.return_value = pd.DataFrame(data, columns=['organisationIdentifierScheme', 'organisationIdentifier', 'organisationName', 'organisationType', 'studyIdentifier', 'organisationAddress'])
  mock_json = mocker.patch("json.load")
  mock_json.return_value = {}
  expected_1 = Code(codeId='Code1', code='code', codeSystem='codesys', codeSystemVersion='3', decode="GBR")
  expected_2 = Code(codeId='Code2', code='code', codeSystem='codesys', codeSystemVersion='3', decode="FRA")
  expected_3 = Code(codeId='Code3', code='code', codeSystem='codesys', codeSystemVersion='3', decode="FRA")
  mock_code = mocker.patch("usdm_excel.iso_3166.ISO3166.code")
  mock_code.side_effect=[expected_1, expected_2, expected_3]
  ids = StudyIdentifiersSheet("")
  assert len(ids.identifiers) == 3
  assert ids.identifiers[0].studyIdentifierId == 'Id_1'
  assert ids.identifiers[0].studyIdentifier == 'NCT12345678'
  assert ids.identifiers[0].studyIdentifierScope.organisationName == 'ClinicalTrials.gov'
  assert ids.identifiers[0].studyIdentifierScope.organizationLegalAddress.city == 'city'
  assert ids.identifiers[1].studyIdentifierId == 'Id_2'
  assert ids.identifiers[1].studyIdentifier == 'NCT12345679'
  assert ids.identifiers[1].studyIdentifierScope.organisationName == 'ClinicalTrials2.gov'
  assert ids.identifiers[1].studyIdentifierScope.organizationLegalAddress.city == 'city2'
  assert ids.identifiers[2].studyIdentifierId == 'Id_3'
  assert ids.identifiers[2].studyIdentifier == 'NCT123456710'
  
def test_create_empty(mocker):
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  data = []
  mock_read = mocker.patch("pandas.read_excel")
  mock_read.return_value = pd.DataFrame(data, columns=['studyIdentifierName', 'studyIdentifierDescription', 'studyIdentifierType'])
  ids = StudyIdentifiersSheet("")
  assert len(ids.identifiers) == 0

def test_read_cell_by_name_error(mocker):
  mock_error = mocker.patch("usdm_excel.errors.errors.Errors.add")
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  data = [['Id 1', 'Id One']]
  mock_read = mocker.patch("pandas.read_excel")
  mock_read.return_value = pd.DataFrame(data, columns=['studyIdentifierName', 'studyIdentifierDescription'])
  ids = StudyIdentifiersSheet("")
  mock_error.assert_called()
  assert mock_error.call_args[0][0] == "studyIdentifiers"
  assert mock_error.call_args[0][1] == None
  assert mock_error.call_args[0][2] == None
  assert mock_error.call_args[0][3] == "Exception ['organisationType'] raised reading sheet."
  
def test_address_error(mocker):
  mock_error = mocker.patch("usdm_excel.errors.errors.Errors.add")
  mock_id = mocker.patch("usdm_excel.id_manager.build_id")
  mock_id.side_effect=['Code_1', 'Org_1', 'Addr_1', 'Id_1']
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  data = [
    [ 'USGOV', 'CT-GOV', 'ClinicalTrials.gov', 'Study Registry', 'NCT12345678', 'line|city|district|state|GBR' ],
  ]
  mock_read = mocker.patch("pandas.read_excel")
  mock_read.return_value = pd.DataFrame(data, columns=['organisationIdentifierScheme', 'organisationIdentifier', 'organisationName', 'organisationType', 'studyIdentifier', 'organisationAddress'])
  mock_json = mocker.patch("json.load")
  mock_json.return_value = {}
  expected_1 = Code(codeId='Code1', code='code', codeSystem='codesys', codeSystemVersion='3', decode="GBR")
  mock_code = mocker.patch("usdm_excel.iso_3166.ISO3166.code")
  mock_code.side_effect=[expected_1]
  ids = StudyIdentifiersSheet("")
  mock_error.assert_called()
  assert mock_error.call_args[0][0] == "studyIdentifiers"
  assert mock_error.call_args[0][1] == 1
  assert mock_error.call_args[0][2] == 6
  assert mock_error.call_args[0][3] == "Address does not contain the required fields (line, city, district, state, postal code and country code) using '|' separator characters, only 5 found"



from typing import List, Union
from .api_base_model import ApiBaseModel

class BiomedicalConceptSurrogate(ApiBaseModel):
  bcSurrogateId: str
  bcSurrogateName: str
  bcSurrogateDescription: Union[str, None] = None
  bcSurrogateReference: str



from usdm_model.organisation import Organisation
from usdm_model.address import Address
from usdm_model.study_identifier import StudyIdentifier
from usdm_excel.base_sheet import BaseSheet
from usdm_excel.id_manager import id_manager
from usdm_excel.iso_3166 import ISO3166
import pandas as pd
import traceback

class StudyIdentifiersSheet(BaseSheet):

  def __init__(self, file_path: str):
    try:
      super().__init__(file_path=file_path, sheet_name='studyIdentifiers')
      self.identifiers = []
      self.process_sheet()
    except Exception as e:
      self._general_error(f"Exception [{e}] raised reading sheet.")
      self._traceback(f"{traceback.format_exc()}")
      
  def process_sheet(self):
    self.identifiers = []
    for index, row in self.sheet.iterrows():
      organisation_type = self.read_cdisc_klass_attribute_cell_by_name('Organisation', 'organisationType', index, 'organisationType')     
      try:
        organisation = Organisation(
          organisationId=id_manager.build_id(Organisation),
          organisationIdentifierScheme=self.read_cell_by_name(index, 'organisationIdentifierScheme'), 
          organisationIdentifier=self.read_cell_by_name(index, 'organisationIdentifier'),
          organisationName=self.read_cell_by_name(index, 'organisationName'),
          organisationType=organisation_type,
          organizationLegalAddress=self._build_address(index)
        )
      except Exception as e:
        self._general_error(f"Failed to create Organisation object, exception {e}")
      else:
        try:
          self.identifiers.append(StudyIdentifier(
            studyIdentifierId=id_manager.build_id(StudyIdentifier),
            studyIdentifier=self.read_cell_by_name(index, 'studyIdentifier'), 
            studyIdentifierScope=organisation)
          )
        except Exception as e:
          self._general_error(f"Failed to create StudyIdentifier object, exception {e}")
    
  def _build_address(self, row_index):
    field_name = 'organisationAddress'
    raw_address = self.read_cell_by_name(row_index, field_name)
    # The '|' separator is preserved for legacy reasons but should be removed in the future
    if '|' in raw_address:
      sep = '|'
      parts = raw_address.split(sep)
    else:
      sep = ','
      parts = self._state_split(raw_address)
    if len(parts) == 6:
      result = self._to_address(
          id_manager.build_id(Address),
          parts[0].strip(), 
          parts[1].strip(), 
          parts[2].strip(), 
          parts[3].strip(), 
          parts[4].strip(), 
          ISO3166().code(parts[5].strip())
        )
      return result
    else:
      col_index = self.sheet.columns.get_loc(field_name)
      self._error(row_index, col_index, f"Address does not contain the required fields (line, city, district, state, postal code and country code) using '{sep}' separator characters, only {len(parts)} found")
      return None

  def _to_address(self, id, line, city, district, state, postal_code, country):
    text = "%s, %s, %s, %s, %s, %s" % (line, city, district, state, postal_code, country.decode)
    text = text.replace(' ,', '')
    try:
      result = Address(addressId=id, text=text, line=line, city=city, district=district, state=state, postalCode=postal_code, country=country)
    except Exception as e:
      self._general_error(f"Failed to create Address object, exception {e}")
      result = None
    return result


from usdm_excel.base_sheet import BaseSheet
from usdm_excel.cross_ref import cross_references
from usdm_excel.id_manager import id_manager
from usdm_model.study_epoch import StudyEpoch
import traceback

class StudyDesignEpochSheet(BaseSheet):

  def __init__(self, file_path):
    try:
      super().__init__(file_path=file_path, sheet_name='studyDesignEpochs')
      self.items = []
      for index, row in self.sheet.iterrows():
        name = self.read_cell_by_name(index, 'studyEpochName')
        description = self.read_description_by_name(index, 'studyEpochDescription')
        epoch_type = self.read_cdisc_klass_attribute_cell_by_name('StudyEpoch', 'studyEpochType', index, 'studyEpochType')
        try:
          item = StudyEpoch(
            studyEpochId=id_manager.build_id(StudyEpoch), 
            studyEpochName=name, 
            studyEpochDescription=description,
            studyEpochType=epoch_type
          )
        except Exception as e:
          self._general_error(f"Failed to create StudyEpoch object, exception {e}")
        else:
          self.items.append(item)
          cross_references.add(name, item)     
    except Exception as e:
      self._general_error(f"Exception [{e}] raised reading sheet.")
      self._traceback(f"{traceback.format_exc()}")


from usdm_excel.base_sheet import BaseSheet
from usdm_excel.cross_ref import cross_references
from usdm_excel.id_manager import id_manager
import traceback
import pandas as pd
from usdm_model.activity import Activity

class StudyDesignActivitySheet(BaseSheet):

  def __init__(self, file_path):
    try:
      super().__init__(file_path=file_path, sheet_name='studyDesignActivities', optional=True)
      self.items = []
      if self.sheet is not None:
        for index, row in self.sheet.iterrows():
          name = self.read_cell_by_name(index, 'activityName')
          description = self.read_description_by_name(index, 'activityDescription')
          conditional = self.read_boolean_cell_by_name(index, 'activityIsConditional')
          reason = self.read_cell_by_name(index, 'activityIsConditionalReason')
          try:
            item = Activity(
              activityId=id_manager.build_id(Activity), 
              activityName=name,
              activityDescription=description,
              activityIsConditional=conditional,
              activityIsConditionalReason=reason
            )
          except Exception as e:
            self._general_error(f"Failed to create Activity object, exception {e}")
          else:
            self.items.append(item)
            cross_references.add(name, item)     
        self.double_link(self.items, 'activityId', 'previousActivityId', 'nextActivityId')   
    except Exception as e:
      self._general_error(f"Exception [{e}] raised reading sheet.")
      self._traceback(f"{traceback.format_exc()}")




from typing import List
from .alias_code import AliasCode
from .api_base_model import ApiBaseModel
from .biomedical_concept_property import BiomedicalConceptProperty

class BiomedicalConcept(ApiBaseModel):
  biomedicalConceptId: str
  bcName: str
  bcSynonyms: List[str] = []
  bcReference: str
  bcProperties: List[BiomedicalConceptProperty] = []
  bcConceptCode: AliasCode



from typing import Union
from .api_base_model import ApiBaseModel
from .code import Code

class StudyEpoch(ApiBaseModel):
  studyEpochId: str
  studyEpochName: str
  studyEpochDescription: Union[str, None] = None
  studyEpochType: Code
  previousStudyEpochId: Union[str, None] = None
  nextStudyEpochId: Union[str, None] = None



from usdm_excel.base_sheet import BaseSheet
from usdm_excel.cross_ref import cross_references
from usdm_excel.id_manager import id_manager
import traceback
import pandas as pd
from usdm_model.procedure import Procedure

class StudyDesignProcedureSheet(BaseSheet):

  def __init__(self, file_path):
    try:
      super().__init__(file_path=file_path, sheet_name='studyDesignProcedures')
      self.procedures = []
      for index, row in self.sheet.iterrows():
        xref = self.read_cell_by_name(index, "xref")
        name = self.read_cell_by_name(index, "procedureName")
        description = self.read_description_by_name(index, 'procedureDescription')
        type = self.read_cell_by_name(index, "procedureType")
        code = self.read_other_code_cell_by_name(index, 'procedureCode')
        conditional = self.read_boolean_cell_by_name(index, 'procedureIsConditional')
        reason = self.read_cell_by_name(index, 'procedureIsConditionalReason')
        try:
          item = Procedure(procedureId=id_manager.build_id(Procedure),
            procedureName=name,
            procedureDescription=description,
            procedureType=type, 
            procedureCode=code, 
            procedureIsConditional=conditional, 
            procedureIsConditionalReason=reason
          )
        except:
          self._general_error(f"Failed to create Procedure object, exception {e}")
        else:
          self.procedures.append(item)
          cross_references.add(xref, item)        
    except Exception as e:
      self._general_error(f"Exception [{e}] raised reading sheet.")
      self._traceback(f"{traceback.format_exc()}")



from typing import Union
from .api_base_model import ApiBaseModel

class IntercurrentEvent(ApiBaseModel):
  intercurrentEventId: str
  intercurrentEventName: str
  intercurrentEventDescription: Union[str, None] = None
  intercurrentEventStrategy: str



import pytest
import pandas as pd

xfail = pytest.mark.xfail

#from src.usdm_excel.base_sheet import BaseSheet
from src.usdm_excel.study_design_arm_sheet.study_design_arm_sheet import StudyDesignArmSheet
from src.usdm_model.code import Code

def test_create(mocker):
  mock_id = mocker.patch("usdm_excel.id_manager.build_id")
  mock_id.side_effect=['ArmId_1', 'ArmId_2', 'ArmId_3']
  expected_1 = Code(codeId='Code1', code='code', codeSystem='codesys', codeSystemVersion='3', decode="label1")
  expected_2 = Code(codeId='Code2', code='code', codeSystem='codesys', codeSystemVersion='3', decode="label2")
  expected_3 = Code(codeId='Code3', code='code', codeSystem='codesys', codeSystemVersion='3', decode="label3")
  expected_4 = Code(codeId='Code4', code='code', codeSystem='codesys', codeSystemVersion='3', decode="label4")
  expected_5 = Code(codeId='Code5', code='code', codeSystem='codesys', codeSystemVersion='3', decode="label5")
  expected_6 = Code(codeId='Code6', code='code', codeSystem='codesys', codeSystemVersion='3', decode="label6")
  mock_code = mocker.patch("usdm_excel.cdisc_ct.CDISCCT.code_for_attribute")
  mock_code.side_effect=[expected_1, expected_2, expected_3, expected_4, expected_5, expected_6]
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  data = [['Arm 1', 'Arm One', 'C12345', 'Subject', 'C99999'], ['Arm 2', 'Arm Two', 'C12345', 'BYOD', 'C99999'], ['Arm 3', 'Arm Three', 'C12345', 'ePRO', 'C99999']]
  mock_read = mocker.patch("pandas.read_excel")
  mock_read.return_value = pd.DataFrame(data, columns=['studyArmName', 'studyArmDescription', 'studyArmType', 'studyArmDataOriginDescription', 'studyArmDataOriginType'])
  epochs = StudyDesignArmSheet("")
  assert len(epochs.items) == 3
  assert epochs.items[0].studyArmId == 'ArmId_1'
  assert epochs.items[0].studyArmName == 'Arm 1'
  assert epochs.items[0].studyArmDataOriginDescription == 'Subject'
  assert epochs.items[0].studyArmDataOriginType == expected_2
  assert epochs.items[1].studyArmId == 'ArmId_2'
  assert epochs.items[1].studyArmDescription == 'Arm Two'
  assert epochs.items[2].studyArmId == 'ArmId_3'
  assert epochs.items[2].studyArmType == expected_5
  assert epochs.items[2].studyArmDataOriginDescription == 'ePRO'
  
def test_create_empty(mocker):
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  data = []
  mock_read = mocker.patch("pandas.read_excel")
  mock_read.return_value = pd.DataFrame(data, columns=['studyArmName', 'studyArmDescription', 'studyArmType'])
  epochs = StudyDesignArmSheet("")
  assert len(epochs.items) == 0

def test_read_cell_by_name_error(mocker):
  mock_error = mocker.patch("usdm_excel.errors.errors.Errors.add")
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  data = [['Arm 1', 'Arm One']]
  mock_read = mocker.patch("pandas.read_excel")
  mock_read.return_value = pd.DataFrame(data, columns=['studyArmName', 'studyArmDescription'])
  epochs = StudyDesignArmSheet("")
  mock_error.assert_called()
  assert mock_error.call_args[0][0] == "studyDesignArms"
  assert mock_error.call_args[0][1] == None
  assert mock_error.call_args[0][2] == None
  assert mock_error.call_args[0][3] == "Exception ['studyArmType'] raised reading sheet."
  



import json
from usdm_excel.id_manager import id_manager
from usdm_excel.configuration_sheet import ConfigurationSheet
from usdm_excel.study_sheet.study_sheet import StudySheet
from usdm_excel.nodes_and_edges import NodesAndEdges
from usdm_excel.cross_ref import cross_references
from usdm_excel.ct_version_manager import ct_version_manager
from usdm_excel.errors.errors import error_manager, Errors
from usdm_excel.option_manager import option_manager as om # Using 'om' as a name clash in pytest
from usdm_excel.cdisc_biomedical_concept import cdisc_bc_library

class USDMExcel():

  FULL_VIEW = NodesAndEdges.FULL
  TIMELINE_VIEW = NodesAndEdges.TIMELINE

  def __init__(self, file_path):
    id_manager.clear()
    cross_references.clear()
    ct_version_manager.clear()
    error_manager.clear()
    om.clear()
    self.configuration = ConfigurationSheet(file_path)
    self.study = StudySheet(file_path)

  def identifier(self):
    study = self.study.the_study()
    if study == None:
      return None
    else:
      return study.study_identifier()

  def the_study(self):
    return self.study.the_study()
  
  def to_json(self):
    try:
      raw_json = self.study.api_root().to_json()
    except Exception as e:
      message = f"Failed to generate JSON output, exception {e}"
      error_manager.add(None, None, None, message)
      raw_json = json.dumps({'error': message}, indent = 2)
    return raw_json
  
  def to_nodes_and_edges(self, view=FULL_VIEW):
    return NodesAndEdges(self.study.the_study(), view).nodes_and_edges()

  def errors(self):
    return error_manager.dump(Errors.WARNING)


import re
from usdm_excel.iso_8601_duration import ISO8601Duration

class TimepointType():

  def __init__(self, parent, row_index, col_index):
    self.__parent = parent
    self.__row = row_index
    self.__col = col_index
    self.timing_type = None
    self.relative_ref = 0
    self.description = None
    self.value = ISO8601Duration.ZERO_DURATION
    timing_info = self.__parent.read_cell(row_index, col_index)
    if timing_info:
      timing_parts = timing_info.split(":")
      if len(timing_parts) == 2:
        if timing_parts[0].upper()[0] == "A":
          self.timing_type = "anchor"
          self.relative_ref = 0
        if timing_parts[0].upper()[0] == "P":
          self.timing_type = "previous"
          self.relative_ref = self._get_relative_ref(timing_parts[0]) * -1
          self._set_text_and_encoded(timing_parts[1])
        elif timing_parts[0].upper()[0] == "N":
          self.timing_type = "next"
          self.relative_ref = self._get_relative_ref(timing_parts[0])
          self._set_text_and_encoded(timing_parts[1])
        elif timing_parts[0].upper()[0] == "C":
          self.timing_type = "cycle start"
          self.relative_ref = self._get_relative_ref(timing_parts[0])
          self._set_text_and_encoded(timing_parts[1])
      else:
        self._log_error(f"Could not decode the timing value, no ':' detected in '{timing_info}'")
    else:
      self._log_error(f"Could not decode the timing value, cell was empty")
 
  def _set_text_and_encoded(self, duration):
    the_duration = duration.strip()
    original_duration = the_duration
    for char in ['+', '-']:
      if char in the_duration:
        the_duration = the_duration.replace(char, "")
        self._log_warning(f"Ignoring '{char}' in {original_duration}")
    duration_parts = re.findall(r"[^\W\d_]+|\d+", the_duration)
    if len(duration_parts) == 2:
      try:
        self.description = original_duration
        self.value = ISO8601Duration().encode(duration_parts[0].strip(), duration_parts[1].strip())
      except Exception as e:
        self._log_error(f"Could not decode the duration value '{the_duration}'")
    else:
      self._log_error(f"Could not decode the duration value, no value and units found in '{the_duration}'")

  def _get_relative_ref(self, part):
    part = part.strip()
    return int(part[1:]) if len(part) > 1 else 1
  
  def _log_error(self, message):
    self.__parent._error(self.__row, self.__col, message)

  def _log_warning(self, message):
    self.__parent._warning(self.__row, self.__col, message)


import pytest
import pandas as pd

xfail = pytest.mark.xfail

#from src.usdm_excel.base_sheet import BaseSheet
from src.usdm_excel.study_design_epoch_sheet.study_design_epoch_sheet import StudyDesignEpochSheet
from src.usdm_model.code import Code

def test_create(mocker):
  mock_id = mocker.patch("usdm_excel.id_manager.build_id")
  mock_id.side_effect=['EpochId_1', 'EpochId_2', 'EpochId_3']
  expected_1 = Code(codeId='Code1', code='code', codeSystem='codesys', codeSystemVersion='3', decode="label1")
  expected_2 = Code(codeId='Code2', code='code', codeSystem='codesys', codeSystemVersion='3', decode="label2")
  expected_3 = Code(codeId='Code3', code='code', codeSystem='codesys', codeSystemVersion='3', decode="label3")
  mock_code = mocker.patch("usdm_excel.cdisc_ct.CDISCCT.code_for_attribute")
  mock_code.side_effect=[expected_1, expected_2, expected_3]
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  data = [['Epoch 1', 'Epoch One', 'C12345'], ['Epoch 2', 'Epoch Two', 'C12345'], ['Epoch 3', 'Epoch Three', 'C12345']]
  mock_read = mocker.patch("pandas.read_excel")
  mock_read.return_value = pd.DataFrame(data, columns=['studyEpochName', 'studyEpochDescription', 'studyEpochType'])
  epochs = StudyDesignEpochSheet("")
  assert len(epochs.items) == 3
  assert epochs.items[0].studyEpochId == 'EpochId_1'
  assert epochs.items[0].studyEpochName == 'Epoch 1'
  assert epochs.items[1].studyEpochId == 'EpochId_2'
  assert epochs.items[1].studyEpochDescription == 'Epoch Two'
  assert epochs.items[2].studyEpochId == 'EpochId_3'
  assert epochs.items[2].studyEpochType == expected_3
  
def test_create_empty(mocker):
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  data = []
  mock_read = mocker.patch("pandas.read_excel")
  mock_read.return_value = pd.DataFrame(data, columns=['studyEpochName', 'studyEpochDescription', 'studyEpochType'])
  epochs = StudyDesignEpochSheet("")
  assert len(epochs.items) == 0

def test_read_cell_by_name_error(mocker):
  mock_error = mocker.patch("usdm_excel.errors.errors.Errors.add")
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  data = [['Epoch 1', 'Epoch One']]
  mock_read = mocker.patch("pandas.read_excel")
  mock_read.return_value = pd.DataFrame(data, columns=['studyEpochName', 'studyEpochDescription'])
  epochs = StudyDesignEpochSheet("")
  mock_error.assert_called()
  assert mock_error.call_args[0][0] == "studyDesignEpochs"
  assert mock_error.call_args[0][1] == None
  assert mock_error.call_args[0][2] == None
  assert mock_error.call_args[0][3] == "Exception ['studyEpochType'] raised reading sheet."
  



from typing import List, Union
from .api_base_model import ApiBaseModel
from .code import Code

class InvestigationalIntervention(ApiBaseModel):
  investigationalInterventionId: str
  interventionDescription: str
  codes: List[Code] = []
